--[[
Interface: 1.7.1.0 b10490

Copyright (C) GtX (Andy), 2018

Author: GtX | Andy
Date: 13.12.2018

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

History:
V 1.0.0.0 @ 13.12.2018 - Release Version for FS19. (Based / Converted from my FS17 'ServiceTrailer.lua'.)
V 1.0.1.0 @ 20.05.2019 - Add support for 'Workshop Tabber 1.0.0.0'
V 1.0.2.0 @ 25.07.2019 - Added support for Seasons Workshop changes.
V 1.0.3.0 @ 11.03.2020 - Fixed update error when placing a 'mission pallet' in service trigger, fix MP event bugs, small optimisations.
V 1.1.0.0 @ 22.11.2020 - More optimisations, added support for electric vehicles from Alpine Pack and also mods using the fillType 'ELECTRICCHARGE'.
V 1.1.1.0 @ 21.04.2020 - Fixed access for contractor farms.

Important:
Free for use in other mods - no permission needed.
No changes are to be made to this script without permission from GtX | Andy

Frei verwendbar - keine erlaubnis nötig
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
]]


ServiceVehicle = {}
ServiceVehicle.MOD_NAME = g_currentModName

ServiceVehicle.MARKER_RED = {1, 0, 0, 1}
ServiceVehicle.MARKER_GREEN = {0, 1, 0, 1}

function ServiceVehicle.prerequisitesPresent(specializations)
    return true
end

function ServiceVehicle.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad",                     ServiceVehicle)
    SpecializationUtil.registerEventListener(vehicleType, "onDelete",                   ServiceVehicle)
    SpecializationUtil.registerEventListener(vehicleType, "onReadStream",               ServiceVehicle)
    SpecializationUtil.registerEventListener(vehicleType, "onWriteStream",              ServiceVehicle)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdate",                   ServiceVehicle)
    SpecializationUtil.registerEventListener(vehicleType, "onFillUnitFillLevelChanged", ServiceVehicle)
end

function ServiceVehicle.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "loadMarkers",                 ServiceVehicle.loadMarkers)
    SpecializationUtil.registerFunction(vehicleType, "load3DText",                  ServiceVehicle.load3DText)
    SpecializationUtil.registerFunction(vehicleType, "setServiceMarker",            ServiceVehicle.setServiceMarker)
    SpecializationUtil.registerFunction(vehicleType, "setServiceFuelState",         ServiceVehicle.setServiceFuelState)
    SpecializationUtil.registerFunction(vehicleType, "getVehicleFuelLevel",         ServiceVehicle.getVehicleFuelLevel)
    SpecializationUtil.registerFunction(vehicleType, "setServiceMarkerState",       ServiceVehicle.setServiceMarkerState)
    SpecializationUtil.registerFunction(vehicleType, "setServiceAnimationState",    ServiceVehicle.setServiceAnimationState)
    SpecializationUtil.registerFunction(vehicleType, "updateServiceInformation",    ServiceVehicle.updateServiceInformation)
    SpecializationUtil.registerFunction(vehicleType, "getCurrentVehicleToService",  ServiceVehicle.getCurrentVehicleToService)
    SpecializationUtil.registerFunction(vehicleType, "setIsCharging",               ServiceVehicle.setIsCharging)
    SpecializationUtil.registerFunction(vehicleType, "setVehicleToCharge",          ServiceVehicle.setVehicleToCharge)
    SpecializationUtil.registerFunction(vehicleType, "getIsChargingVehicle",        ServiceVehicle.getIsChargingVehicle)
    SpecializationUtil.registerFunction(vehicleType, "onElectricVehicleDeleted",    ServiceVehicle.onElectricVehicleDeleted)
end

function ServiceVehicle:onLoad(savegame)
    self.spec_serviceVehicle = self[string.format("spec_%s.serviceVehicle", ServiceVehicle.MOD_NAME)]

    if self.spec_serviceVehicle == nil then
        g_logManager:error("[%s] Specialization with name 'serviceVehicle' was not found in modDesc!", ServiceVehicle.MOD_NAME)
    end

    local spec = self.spec_serviceVehicle
    local triggers = {
        playerTrigger = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.serviceVehicle.service#playerTriggerNode"), self.i3dMappings),
        sellTriggerNode = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.serviceVehicle.service#sellTriggerNode"), self.i3dMappings)
    }

    spec.texts = {
        activateService = ServiceVehicle.returnText("action_activateService", "Activate Service Function", self.customEnvironment),
        deactivateService = ServiceVehicle.returnText("action_deactivateService", "Deactivate Service Function", self.customEnvironment)
    }

    spec.isChargingVehicle = false
    spec.electricVehiclesInRange = {}
    spec.hasElectricCharge = FillType.ELECTRICCHARGE ~= nil and Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.serviceVehicle.fuel#allowElectricCharge"), false)

    spec.consumerFillTypeToTitle = {
        [FillType.DIESEL] = string.upper(g_i18n:getText("fillType_diesel")) .. " :"
    }

    if spec.hasElectricCharge then
        spec.consumerFillTypeToTitle[FillType.ELECTRICCHARGE] = "BATTERY :"
        spec.texts.batteryChargeTime = ServiceVehicle.returnText("info_chargeTime", "Time until fully charged: %d Minutes %d Seconds (%d%%)", "pdlc_alpineFarmingPack")
    end

    -- Only the triggers are required for the script to work. All other features are optional.
    -- l10n entries 'action_activateService' and 'action_deactivateService' are also a minimum requirement. (No inputBinding is needed.)
    if triggers.playerTrigger ~= nil and triggers.sellTriggerNode ~= nil then
        spec.serviceActivatable = ServiceVehicleActivatable:new(self, triggers)

        if spec.serviceActivatable ~= nil then
            if self.spec_animatedVehicle ~= nil then
                local animationName = getXMLString(self.xmlFile, "vehicle.serviceVehicle.service.animation#name")

                if animationName ~= nil and self:getAnimationExists(animationName) then
                    spec.texts.activateServiceMenu = ServiceVehicle.returnText("action_activateServiceMenu", "Service Menu", self.customEnvironment)

                    spec.serviceAnimation = {
                        name = animationName,
                        isActive = false,
                        finished = false,
                        markers = self:loadMarkers("vehicle.serviceVehicle.service.animation.markers", savegame, "serviceActive")
                    }

                    spec.serviceActivatable.hasExtraInputs = true
                    spec.serviceActivatable.hasMarkers = spec.serviceAnimation.markers ~= nil
                end
            end

            if self.isClient and hasXMLProperty(self.xmlFile, "vehicle.serviceVehicle.service.textRender3D") then
                local keyToText = {
                    ["mainHeader"] = string.upper(g_i18n:getText("ui_garageOverview")),
                    ["vehicleOwner"] = string.upper(g_i18n:getText("ui_farmName")) .. " :",
                    ["vehicleName"] = string.upper(g_i18n:getText("ui_vehicle")) .. " :",
                    ["vehicleOperatingTime"] = string.upper(g_i18n:getText("ui_operatingHours")) .. " :",
                    ["vehicleAge"] = string.upper(g_i18n:getText("ui_age")) .. " :",
                    ["vehicleFuel"] = string.upper(g_i18n:getText("configuration_consumer")) .. " :",
                    ["vehicleCondition"] = string.upper(g_i18n:getText("ui_condition")) .. " :"
                }

                for name, text in pairs (keyToText) do
                    self:load3DText("vehicle.serviceVehicle.service.textRender3D.headers", name, text, true, "headers", "numHeaders")

                    if name ~= "mainHeader" then
                        self:load3DText("vehicle.serviceVehicle.service.textRender3D.texts", name, "", false, "texts", "numTexts")
                    end
                end
            end
        end
    else
        if triggers.playerTrigger == nil then
            g_logManager:error("[ServiceVehicle] - Service Player Trigger could not be found in '%s', This is a minimum requirement!",  self.configFileName)
        end

        if triggers.sellTriggerNode == nil then
            g_logManager:error("[ServiceVehicle] - Vehicle Service Trigger could not be found in '%s', This is a minimum requirement!",  self.configFileName)
        end
    end

    if self.spec_fillTriggerVehicle ~= nil then
        local fillTrigger = self.spec_fillTriggerVehicle.fillTrigger
        local fillUnitIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.serviceVehicle.fuel#fillUnitIndex"), 1)

        if self:getFillUnitSupportsFillType(fillUnitIndex, FillType.DIESEL) then
            local playerTrigger = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.serviceVehicle.fuel#playerTriggerNode"), self.i3dMappings)

            if playerTrigger ~= nil then
                spec.texts.diesel = g_i18n:getText("fillType_diesel", "Diesel")
                spec.texts.startFuelPump = ServiceVehicle.returnText("action_startFuelPump", "Start Fuel Pump", self.customEnvironment)
                spec.texts.stopFuelPump = ServiceVehicle.returnText("action_stopFuelPump", "Stop Fuel Pump", self.customEnvironment)
                spec.texts.fuelTankEmpty = ServiceVehicle.returnText("warning_fuelTankEmpty", "First Refill Fuel Tanker!", self.customEnvironment)

                spec.fuelActivatable = ServiceVehicleFuelActivatable:new(self, playerTrigger)

                if spec.fuelActivatable ~= nil then
                    spec.fuel = {
                        fillUnitIndex = fillUnitIndex,
                        isActive = false,
                        markers = self:loadMarkers("vehicle.serviceVehicle.fuel.markers", savegame, "fuelActive")
                    }

                    spec.fuelActivatable.hasMarkers = spec.fuel.markers ~= nil

                    spec.fuel.usagePerMS = math.max(Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.serviceVehicle.fuel#dieselUsagePerSecond"), 0), 0)

                    if spec.fuel.usagePerMS > 0 then
                        spec.fuel.usagePerMS = math.max(spec.fuel.usagePerMS, 0.01) / 1000
                    end

                    spec.fuel.chargingUsagePerMS = math.max(Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.serviceVehicle.fuel.electricCharge#dieselUsagePerSecond"), 0.01), 0.01) / 1000

                    if self.spec_animatedVehicle ~= nil then
                        local animationName = getXMLString(self.xmlFile, "vehicle.serviceVehicle.fuel.animation#name")

                        if animationName ~= nil and self:getAnimationExists(animationName) then
                            spec.fuelAnimation = {
                                name = animationName
                            }
                        end
                    end

                    if self.isClient then
                        spec.fuel.animationNodes = g_animationManager:loadAnimations(self.xmlFile, "vehicle.serviceVehicle.fuel.animationNodes", self.components, self, self.i3dMappings)

                        spec.fuel.samples = {
                            pumpStart = g_soundManager:loadSampleFromXML(self.xmlFile, "vehicle.serviceVehicle.fuel.sounds", "pumpStart", self.baseDirectory, self.components, 1, AudioGroup.VEHICLE, self.i3dMappings, self),
                            pumpStop  = g_soundManager:loadSampleFromXML(self.xmlFile, "vehicle.serviceVehicle.fuel.sounds", "pumpStop", self.baseDirectory, self.components, 1, AudioGroup.VEHICLE, self.i3dMappings, self),
                            pumpWork  = g_soundManager:loadSampleFromXML(self.xmlFile, "vehicle.serviceVehicle.fuel.sounds", "pumpWork", self.baseDirectory, self.components, 0, AudioGroup.VEHICLE, self.i3dMappings, self)
                        }

                        local exhaustKey = "vehicle.serviceVehicle.fuel.exhaustParticles"

                        if hasXMLProperty(self.xmlFile, exhaustKey) then
                            spec.fuel.exhaustParticles = {}

                            local linkNode = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, exhaustKey .. "#linkNode"), self.i3dMappings)
                            ParticleUtil.loadParticleSystem(self.xmlFile, spec.fuel.exhaustParticles, exhaustKey, self.components, false, nil, self.baseDirectory, linkNode)
                        end
                    end

                    if spec.hasElectricCharge then
                        spec.chargingActivatable = ServiceVehicleChargingActivatable:new(self)

                        if spec.chargingActivatable ~= nil then
                            spec.electricChargePerMS = (getXMLFloat(self.xmlFile, "vehicle.serviceVehicle.fuel.electricCharge#fillLitersPerSecond") or 0.05) / 1000
                        else
                            spec.hasElectricCharge = false
                        end
                    end

                    fillTrigger.fillTriggerCallback = Utils.overwrittenFunction(fillTrigger.fillTriggerCallback, function(trigger, superFunc, triggerId, otherId, onEnter, onLeave, onStay)
                        if trigger.isEnabled and (onEnter or onLeave) then
                            local vehicle = g_currentMission:getNodeObject(otherId)

                            if vehicle ~= nil and vehicle.addFillUnitTrigger ~= nil and vehicle.removeFillUnitTrigger ~= nil and vehicle ~= trigger and vehicle ~= trigger.sourceObject then
                                local fillType = trigger:getCurrentFillType()
                                local fillUnitIndex = vehicle:getFirstValidFillUnitToFill(fillType)

                                if fillUnitIndex ~= nil then
                                    if onEnter then
                                        if trigger.vehiclesTriggerCount[vehicle] == nil then
                                            trigger.vehiclesTriggerCount[vehicle] = 1
                                            vehicle:addFillUnitTrigger(trigger, fillType, fillUnitIndex)
                                        end
                                    else
                                        trigger.vehiclesTriggerCount[vehicle] = nil
                                        vehicle:removeFillUnitTrigger(trigger)
                                    end
                                else
                                    if trigger.sourceObject.setVehicleToCharge ~= nil then
                                        trigger.sourceObject:setVehicleToCharge(vehicle, onEnter)
                                    end
                                end
                            end
                        end
                    end)

                    -- Fix issue in base 'fillVehicle' function not allowing 'fillLitersPerSecond' to be used and fill sound not stopping if vehicle is running as 'delta = 0' was never reached.
                    fillTrigger.fillVehicle = Utils.overwrittenFunction(fillTrigger.fillVehicle, function(trigger, superFunc, vehicle, delta, dt)
                        if trigger.fillLitersPerSecond ~= nil then
                            delta = math.max(math.min(delta, trigger.fillLitersPerSecond * 0.001 * dt), 0.009)
                        end

                        local farmId = vehicle:getActiveFarm()

                        if trigger.sourceObject ~= nil then
                            local sourceFuelFillLevel = trigger.sourceObject:getFillUnitFillLevel(trigger.fillUnitIndex) or 0

                            if sourceFuelFillLevel > 0 and g_currentMission.accessHandler:canFarmAccess(farmId, trigger.sourceObject) then
                                delta = math.min(delta, sourceFuelFillLevel)

                                if delta <= 0 then
                                    return 0
                                end
                            else
                                return 0
                            end
                        end

                        local fillType = trigger:getCurrentFillType()
                        local fillUnitIndex = vehicle:getFirstValidFillUnitToFill(fillType)

                        if fillUnitIndex == nil then
                            return 0
                        end

                        delta = vehicle:addFillUnitFillLevel(farmId, fillUnitIndex, delta, fillType, ToolType.TRIGGER, nil)

                        if delta > 0 then
                            if trigger.sourceObject ~= nil then
                                trigger.sourceObject:addFillUnitFillLevel(farmId, trigger.fillUnitIndex, -delta, fillType, ToolType.TRIGGER, nil)
                            end
                        end

                        if delta < 0.009 then
                            delta = 0
                        end

                        return delta
                    end)

                    local chargingChangeObjects = {}
                    ObjectChangeUtil.loadObjectChangeFromXML(self.xmlFile, "vehicle.serviceVehicle.fuel.electricCharge", chargingChangeObjects, self.components, self)
                    ObjectChangeUtil.setObjectChanges(chargingChangeObjects, spec.hasElectricCharge, nil, nil)
                end
            else
                g_logManager:error("[ServiceVehicle] - Fuel Player Trigger could not be found in '%s'! This is a minimum requirement.",  self.configFileName)
            end
        else
            g_logManager:error("[ServiceVehicle] - DIESEL fillType could not be found at fillUnitIndex '%d' in '%s'!",  fillUnitIndex, self.configFileName)
        end
    end
end

function ServiceVehicle:loadMarkers(baseKey, savegame, savegameKey)
    local markers
    local groupNode = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, baseKey .. "#groupNode"), self.i3dMappings)

    if groupNode ~= nil then
        local spec = self.spec_serviceVehicle

        markers = {
            showMarkers = true,
            groupNode = groupNode,
            useShader = Utils.getNoNil(getXMLBool(self.xmlFile, baseKey .. "#useEmissiveFalloffShader"), false)
        }

        if markers.useShader then
            markers.active = false
            markers.numMarkers = 0
            markers.vehiclesInTrigger = false

            local numMarkers = getNumOfChildren(groupNode)

            if numMarkers > 0 then
                markers.group = {}

                for i = 0, numMarkers - 1 do
                    markers.numMarkers = markers.numMarkers + 1
                    markers.group[markers.numMarkers] = getChildAt(groupNode, i)
                end
            end
        end

        if spec.texts.triggerMarkers == nil then
            spec.texts.triggerMarkers = ServiceVehicle.returnText("action_triggerMarkers", "Trigger Marker", self.customEnvironment)
        end

        setVisibility(markers.groupNode, false)

        if savegame ~= nil and not savegame.resetVehicles then
            local triggerMarkerKey = string.format("%s.%s.serviceVehicle.triggerMarkers#%s", savegame.key, self.customEnvironment, savegameKey)
            markers.showMarkers = Utils.getNoNil(getXMLBool(savegame.xmlFile, triggerMarkerKey), markers.showMarkers)
        end
    end

    return markers
end

function ServiceVehicle:load3DText(baseKey, key, backupText, isHeader, mainKey, numName)
    local spec = self.spec_serviceVehicle

    if spec ~= nil then
        local textsKey = baseKey .. "." .. key

        local serviceInformation = {
            keyName = key,
            text = backupText,
            node = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, textsKey .. "#node"), self.i3dMappings)
        }

        if isHeader then
            local text = getXMLString(self.xmlFile, textsKey .. "#text")

            if text ~= nil then
                serviceInformation.text = g_i18n:getText(text)
            end

            if key ~= "mainHeader" then
                key = key .. "Header"
                serviceInformation.keyName = key
            end
        end

        serviceInformation.size = Utils.getNoNil(getXMLFloat(self.xmlFile, textsKey .. "#textSize"), 0.04)
        serviceInformation.colour = Utils.getNoNil(StringUtil.getVectorNFromString(getXMLString(self.xmlFile, textsKey .. "#textColour"), 4), {0.843, 0.745, 0.705, 1.0})

        local alignment = string.upper(Utils.getNoNil(getXMLString(self.xmlFile, textsKey .. "#textAlignment"), "LEFT"))

        if alignment == "RIGHT" then
            serviceInformation.alignment = RenderText.ALIGN_RIGHT
        elseif alignment == "CENTER" then
            serviceInformation.alignment = RenderText.ALIGN_CENTER
        else
            serviceInformation.alignment = RenderText.ALIGN_LEFT
        end

        if spec.serviceInformation == nil then
            spec.serviceInformation = {
                textsByName = {},
                nameToNode = {},
                numHeaders = 0,
                numTexts = 0
            }
        end

        spec.serviceInformation.textsByName[key] = serviceInformation
        spec.serviceInformation.nameToNode[key] = serviceInformation.node

        if serviceInformation.node ~= nil then
            if spec.serviceInformation[mainKey] == nil then
                spec.serviceInformation[mainKey] = {}
            end

            table.insert(spec.serviceInformation[mainKey], serviceInformation)
            spec.serviceInformation[numName] = #spec.serviceInformation[mainKey]
        end
    end
end

function ServiceVehicle:onDelete()
    local spec = self.spec_serviceVehicle

    if spec.serviceActivatable ~= nil then
        spec.serviceActivatable:delete()
    end

    if spec.fuel ~= nil then
        if spec.fuelActivatable ~= nil then
            spec.fuelActivatable:delete()
        end

        if spec.chargingActivatable ~= nil then
            spec.chargingActivatable:delete()
            spec.chargingActivatable = nil
        end

        if spec.hasElectricCharge then
            self:setIsCharging(false)

            for vehicle, _ in pairs (spec.electricVehiclesInRange) do
                if vehicle.removeDeleteListener ~= nil then
                    vehicle:removeDeleteListener(self, "onElectricVehicleDeleted")
                end
            end

            spec.electricVehiclesInRange = {}
            spec.hasElectricCharge = false
        end

        if self.isClient then
            g_soundManager:deleteSamples(spec.fuel.samples)
            ParticleUtil.deleteParticleSystem(spec.fuel.exhaustParticles)
            g_animationManager:deleteAnimations(spec.fuel.animationNodes)
        end
    end
end

function ServiceVehicle:saveToXMLFile(xmlFile, key, usedModNames)
    local spec = self.spec_serviceVehicle

    if spec.serviceAnimation ~= nil and spec.serviceAnimation.markers ~= nil then
        setXMLBool(xmlFile, key .. ".triggerMarkers#serviceActive", spec.serviceAnimation.markers.showMarkers)
    end

    if spec.fuel ~= nil and spec.fuel.markers ~= nil then
        setXMLBool(xmlFile, key .. ".triggerMarkers#fuelActive", spec.fuel.markers.showMarkers)
    end
end

function ServiceVehicle:onReadStream(streamId, connection)
    local spec = self.spec_serviceVehicle

    if spec.serviceAnimation ~= nil then
        local isActive = streamReadBool(streamId)
        self:setServiceAnimationState(isActive, true)

        if spec.serviceAnimation.markers ~= nil then
            local showMarkers = streamReadBool(streamId)
            self:setServiceMarkerState(showMarkers, true, true)
        end
    end

    if spec.fuel ~= nil then
        local isActive = streamReadBool(streamId)
        self:setServiceFuelState(isActive, true)

        if spec.fuel.markers ~= nil then
            local showMarkers = streamReadBool(streamId)
            self:setServiceMarkerState(showMarkers, false, true)
        end
    end

    if spec.hasElectricCharge then
        if streamReadBool(streamId) then
            local targetVehicle = NetworkUtil.readNodeObject(streamId)
            local fillUnitIndex = streamReadUInt8(streamId)

            self:setIsCharging(true, targetVehicle, fillUnitIndex, true)
        end
    end
end

function ServiceVehicle:onWriteStream(streamId, connection)
    local spec = self.spec_serviceVehicle

    if spec.serviceAnimation ~= nil then
        streamWriteBool(streamId, spec.serviceAnimation.isActive)

        if spec.serviceAnimation.markers ~= nil then
            local markersActive = spec.serviceAnimation.markers.showMarkers
            streamWriteBool(streamId, markersActive)
        end
    end

    if spec.fuel ~= nil then
        streamWriteBool(streamId, spec.fuel.isActive)

        if spec.fuel.markers ~= nil then
            local markersActive = spec.fuel.markers.showMarkers
            streamWriteBool(streamId, markersActive)
        end
    end

    if spec.hasElectricCharge then
        if streamWriteBool(streamId, spec.isChargingVehicle) then
            NetworkUtil.writeNodeObject(streamId, spec.currentElectricVehicle)
            streamWriteUInt8(streamId, spec.currentElectricFillUnitIndex)
        end
    end
end

function ServiceVehicle:onUpdate(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    local raiseActive = false
    local spec = self.spec_serviceVehicle

    if spec.fuel ~= nil and spec.fuelActivatable ~= nil then
        if self.isClient then
            if spec.electricVehicleInTrigger then
                if spec.isChargingVehicle and spec.currentElectricVehicle ~= nil and spec.currentElectricFillUnitIndex ~= nil then
                    if g_currentMission.controlledVehicle ~= nil and g_currentMission.controlledVehicle == spec.currentElectricVehicle or spec.fuelActivatable.objectActivated then
                        local fillUnit = spec.currentElectricVehicle:getFillUnitByIndex(spec.currentElectricFillUnitIndex)

                        if fillUnit ~= nil then
                            local chargeTime = (fillUnit.capacity - fillUnit.fillLevel) / (spec.electricChargePerMS * 1000)

                            if chargeTime >= 1 then
                                local minutes = math.floor(chargeTime / 60)
                                g_currentMission:addExtraPrintText(string.format(spec.texts.batteryChargeTime, minutes, chargeTime - (minutes * 60), fillUnit.fillLevel / fillUnit.capacity * 100))
                            end
                        end
                    end
                end
            end

            if spec.fuel.markers ~= nil then
                local vehiclesInTrigger = spec.electricVehicleInTrigger or next(self.spec_fillTriggerVehicle.fillTrigger.vehiclesTriggerCount) ~= nil
                raiseActive = self:setServiceMarker(spec.fuel.markers, vehiclesInTrigger, spec.fuel.isActive, false, raiseActive)
            end
        end

        if self.isServer and spec.fuel.isActive then
            local farmId = self:getActiveFarm()
            local fuelUsagePerMS = spec.fuel.usagePerMS

            if spec.isChargingVehicle then
                if spec.currentElectricVehicle ~= nil then
                    if spec.currentElectricFillUnitIndex == nil then
                        spec.currentElectricFillUnitIndex = spec.currentElectricVehicle:getFirstValidFillUnitToFill(FillType.ELECTRICCHARGE)
                    end

                    local usedFillLevel = spec.currentElectricVehicle:addFillUnitFillLevel(farmId, spec.currentElectricFillUnitIndex, spec.electricChargePerMS * dt, FillType.ELECTRICCHARGE, ToolType.UNDEFINED, nil)

                    if usedFillLevel == nil or usedFillLevel < 0.001 then
                        self:setIsCharging(false)
                    end

                    fuelUsagePerMS = fuelUsagePerMS + spec.fuel.chargingUsagePerMS
                else
                    self:setIsCharging(false)
                end
            end

            if fuelUsagePerMS > 0 then
                self:addFillUnitFillLevel(farmId, spec.fuel.fillUnitIndex, -fuelUsagePerMS * dt, FillType.DIESEL, ToolType.UNDEFINED, nil)
            end

            raiseActive = true
        end
    end

    if self.isClient and spec.serviceActivatable ~= nil then
        local currentVehicle = self:getCurrentVehicleToService(spec.serviceActivatable)

        if spec.serviceAnimation ~= nil and spec.serviceAnimation.markers ~= nil then
            spec.serviceAnimation.finished = self:getAnimationTime(spec.serviceAnimation.name) == 1

            if spec.serviceAnimation.finished then
                raiseActive = self:setServiceMarker(spec.serviceAnimation.markers, currentVehicle ~= nil, spec.serviceAnimation.isActive, false, raiseActive)
            else
                raiseActive = self:setServiceMarker(spec.serviceAnimation.markers, currentVehicle ~= nil, false, spec.serviceAnimation.isActive, raiseActive)
            end
        end

        -- Update headers and vehicle texts
        if spec.serviceInformation ~= nil then
            raiseActive = self:updateServiceInformation(currentVehicle, spec.serviceInformation, raiseActive)
        end
    end

    if raiseActive then
        self:raiseActive()
    end
end

function ServiceVehicle:onFillUnitFillLevelChanged(fillUnitIndex, fillLevelDelta, fillTypeIndex, toolType, fillPositionData, appliedDelta)
    local spec = self.spec_serviceVehicle

    if spec.fuel ~= nil then
        if spec.fuel.isActive and spec.fuel.fillUnitIndex == fillUnitIndex then
            if self:getFillUnitFillLevel(fillUnitIndex) <= 0 then
                self:setServiceFuelState(false, true)
            end
        end
    end
end

function ServiceVehicle:setIsCharging(isCharging, targetVehicle, fillUnitIndex, noEventSend)
    local spec = self.spec_serviceVehicle

    if spec.hasElectricCharge and spec.chargingActivatable ~= nil then
        ServiceVehicleSetIsChargingEvent.sendEvent(self, isCharging, targetVehicle, fillUnitIndex, noEventSend)

        spec.isChargingVehicle = isCharging
        spec.currentElectricVehicle = targetVehicle
        spec.currentElectricFillUnitIndex = fillUnitIndex

        if not isCharging and next(spec.electricVehiclesInRange) == nil then
            spec.electricVehicleInTrigger = false
            g_currentMission:removeActivatableObject(spec.chargingActivatable)
        end

        self:raiseActive()
    end
end

function ServiceVehicle:onElectricVehicleDeleted(vehicle)
    self:setVehicleToCharge(vehicle, false)
end

function ServiceVehicle:setVehicleToCharge(targetVehicle, onEnter)
    local spec = self.spec_serviceVehicle

    if spec.hasElectricCharge and spec.chargingActivatable ~= nil then
        if onEnter then
            if spec.electricVehiclesInRange[targetVehicle] == nil then
                local fillUnitIndex = targetVehicle:getFirstValidFillUnitToFill(FillType.ELECTRICCHARGE)

                if fillUnitIndex ~= nil then
                    spec.electricVehiclesInRange[targetVehicle] = fillUnitIndex

                    if targetVehicle.addDeleteListener ~= nil then
                        targetVehicle:addDeleteListener(self, "onElectricVehicleDeleted")
                    end

                    spec.electricVehicleInTrigger = true
                    g_currentMission:addActivatableObject(spec.chargingActivatable)
                end
            end
        else
            if spec.electricVehiclesInRange[targetVehicle] ~= nil then
                spec.electricVehiclesInRange[targetVehicle] = nil

                if targetVehicle.removeDeleteListener ~= nil then
                    targetVehicle:removeDeleteListener(self, "onElectricVehicleDeleted")
                end

                if self.isServer and spec.currentElectricVehicle == targetVehicle then
                    self:setIsCharging(false)
                end
            end

            if next(spec.electricVehiclesInRange) == nil then
                spec.electricVehicleInTrigger = false
                g_currentMission:removeActivatableObject(spec.chargingActivatable)
            end
        end

        self:raiseActive()
    end
end

function ServiceVehicle:getIsChargingVehicle()
    if self.spec_serviceVehicle.isChargingVehicle then
        return true
    end

    return false
end

function ServiceVehicle:getCurrentVehicleToService(activatable)
    if g_seasons ~= nil and g_directSellDialog ~= nil then
        if activatable.playerInTrigger == g_currentMission.player then
            if ServiceVehicle.getIsVehicleValid(self, g_directSellDialog.vehicle) then
                return g_directSellDialog.vehicle
            end
        end
    end

    return activatable.currentVehicle
end

function ServiceVehicle:getVehicleFuelLevel(vehicle)
    local spec = self.spec_serviceVehicle
    local fillLevelString, consumerTypeTitle, volumeUnit = "", "", "l"

    if vehicle ~= nil and vehicle.spec_motorized ~= nil and vehicle.spec_motorized.consumersByFillTypeName ~= nil then
        local consumer = vehicle.spec_motorized.consumersByFillTypeName.diesel

        if consumer == nil and spec.hasElectricCharge then
            consumer = vehicle.spec_motorized.consumersByFillTypeName.electricCharge
            volumeUnit = "kWh"
        end

        if consumer ~= nil and consumer.fillUnitIndex ~= nil and consumer.fillType ~= nil then
            fillLevelString = string.format("%s %s", g_i18n:formatNumber(vehicle:getFillUnitFillLevel(consumer.fillUnitIndex) or 0, 2), volumeUnit)
            consumerTypeTitle = spec.consumerFillTypeToTitle[consumer.fillType] or ""
        end
    end

    return fillLevelString, consumerTypeTitle
end

function ServiceVehicle:updateServiceInformation(currentVehicle, serviceInfo, raiseActive)
    if currentVehicle ~= nil then
        -- Collect current vehicle information each loop as it changes
        local farmName = "N/A"

        if currentVehicle.getActiveFarm ~= nil then
            local farm = g_farmManager:getFarmById(currentVehicle:getActiveFarm())

            if farm ~= nil and (farm.name ~= nil and farm.name ~= "") then
                farmName = farm.name
            end
        end

        serviceInfo.textsByName.vehicleOwner.text = farmName

        local vehicleName = "N/A"

        if currentVehicle.getFullName ~= nil then
            vehicleName = currentVehicle:getFullName()
        end

        serviceInfo.textsByName.vehicleName.text = vehicleName

        local minutes = (currentVehicle.operatingTime or 0) / (1000 * 60)
        local hours = math.floor(minutes / 60)
        minutes = math.floor((minutes - hours * 60) / 6)
        serviceInfo.textsByName.vehicleOperatingTime.text = string.format("%02d.%01d h", hours, minutes)

        if serviceInfo.nameToNode.vehicleFuel ~= nil then
            local consumersLevel, consumerTypeTitle = self:getVehicleFuelLevel(currentVehicle)

            if consumerTypeTitle ~= "" then
                if serviceInfo.textsByName.vehicleFuelHeader ~= nil then
                    serviceInfo.textsByName.vehicleFuelHeader.text = consumerTypeTitle
                end
            else
                consumersLevel = "N/A"
            end

            serviceInfo.textsByName.vehicleFuel.text = consumersLevel
        end

        serviceInfo.textsByName.vehicleAge.text = string.format(g_i18n:getText("shop_age"), (currentVehicle.age or 0))

        local damagedPercent = 0

        if currentVehicle.getWearTotalAmount ~= nil then
            damagedPercent = math.floor(math.min(currentVehicle:getWearTotalAmount() * 100, 100))
        end

        serviceInfo.textsByName.vehicleCondition.text = tostring(100 - damagedPercent) .. " %"

        -- Render Headers
        for i = 1, serviceInfo.numHeaders do
            ServiceVehicle.drawServiceInformation(self, serviceInfo.headers[i])
        end

        -- Render Vehicle Texts
        for i = 1, serviceInfo.numTexts do
            ServiceVehicle.drawServiceInformation(self, serviceInfo.texts[i])
        end

        -- Reset once at the end for mods that do not use text drawing correctly.
        setTextColor(1, 1, 1, 1)
        setTextAlignment(RenderText.ALIGN_LEFT)

        raiseActive = true
    end

    return raiseActive
end

function ServiceVehicle:setServiceMarker(markers, vehiclesInTrigger, isActive, forceActive, raiseActive)
    if markers.showMarkers then
        if isActive ~= markers.active then
            setVisibility(markers.groupNode, isActive)
            markers.active = isActive
        end

        if markers.useShader then
            if markers.vehiclesInTrigger ~= vehiclesInTrigger then
                markers.vehiclesInTrigger = vehiclesInTrigger

                local colour = ServiceVehicle.MARKER_RED

                if vehiclesInTrigger then
                    colour = ServiceVehicle.MARKER_GREEN
                end

                for i = 1, markers.numMarkers do
                    setShaderParameter(markers.group[i], "colorScale", colour[1], colour[2], colour[3], colour[4], false)
                end
            end

            if isActive or forceActive then
                raiseActive = true
            end
        else
            raiseActive = forceActive
        end
    else
        if markers.active then
            setVisibility(markers.groupNode, false)
            markers.active = false
        end
    end

    return raiseActive
end

function ServiceVehicle:setServiceAnimationState(isActive, noEventSend)
    ServiceVehicleAnimationEvent.sendEvent(self, isActive, noEventSend)

    local spec = self.spec_serviceVehicle

    if isActive ~= spec.serviceAnimation.isActive then
        spec.serviceAnimation.isActive = isActive

        local animationTime = self:getAnimationTime(spec.serviceAnimation.name)
        if isActive and animationTime < 1 then
            self:playAnimation(spec.serviceAnimation.name, 1, animationTime, true)
        elseif isActive and animationTime >= 1 then
            self:playAnimation(spec.serviceAnimation.name, 1, 0, true)
        elseif animationTime > 0 then
            self:playAnimation(spec.serviceAnimation.name, -1, animationTime, true)
        end

        self:raiseActive()
    end
end

function ServiceVehicle:setServiceMarkerState(showMarkers, isService, noEventSend)
    ServiceVehicleMarkerEvent.sendEvent(self, showMarkers, isService, noEventSend)

    local spec = self.spec_serviceVehicle

    if isService then
        spec.serviceAnimation.markers.showMarkers = showMarkers
    else
        spec.fuel.markers.showMarkers = showMarkers
    end

    self:raiseActive()
end

function ServiceVehicle:setServiceFuelState(isActive, noEventSend)
    ServiceVehicleFuelEvent.sendEvent(self, isActive, noEventSend)

    local spec = self.spec_serviceVehicle
    if isActive ~= spec.fuel.isActive then
        spec.fuel.isActive = isActive

        if spec.fuelAnimation ~= nil then
            local animationTime = self:getAnimationTime(spec.fuelAnimation.name)

            if isActive and animationTime < 1 then
                self:playAnimation(spec.fuelAnimation.name, 1, animationTime, true)
            elseif isActive and animationTime >= 1 then
                self:playAnimation(spec.fuelAnimation.name, 1, 0, true)
            elseif animationTime > 0 then
                self:playAnimation(spec.fuelAnimation.name, -1, animationTime, true)
            end
        end

        if self.isClient then
            if isActive then
                if spec.fuel.samples ~= nil then
                    g_soundManager:stopSample(spec.fuel.samples.pumpStop)
                    g_soundManager:playSample(spec.fuel.samples.pumpStart)
                    g_soundManager:playSample(spec.fuel.samples.pumpWork, 0, spec.fuel.samples.pumpStart)
                end

                if spec.fuel.animationNodes ~= nil then
                    g_animationManager:startAnimations(spec.fuel.animationNodes)
                end

                if spec.fuel.exhaustParticles ~= nil then
                    ParticleUtil.setEmittingState(spec.fuel.exhaustParticles, true)
                end
            else
                if spec.fuel.samples ~= nil then
                    g_soundManager:stopSample(spec.fuel.samples.pumpStart)
                    g_soundManager:stopSample(spec.fuel.samples.pumpWork)
                    g_soundManager:playSample(spec.fuel.samples.pumpStop)
                end

                if spec.fuel.animationNodes ~= nil then
                    g_animationManager:stopAnimations(spec.fuel.animationNodes)
                end

                if spec.fuel.exhaustParticles ~= nil then
                    ParticleUtil.setEmittingState(spec.fuel.exhaustParticles, false)
                end
            end
        end

        self:raiseActive()
    end
end

function ServiceVehicle.returnText(i18n, backupText, customEnvironment)
    if g_i18n:hasText(i18n, customEnvironment) then
        return g_i18n:getText(i18n, customEnvironment)
    end

    return backupText
end

function ServiceVehicle.drawServiceInformation(vehicle, serviceInfo)
    if serviceInfo.node ~= nil then
        local x, y, z = getWorldTranslation(serviceInfo.node)
        local rx, ry, rz = getWorldRotation(serviceInfo.node)

        setTextAlignment(serviceInfo.alignment)
        setTextColor(serviceInfo.colour[1], serviceInfo.colour[2], serviceInfo.colour[3], serviceInfo.colour[4])

        renderText3D(x,y,z, rx,ry,rz, serviceInfo.size, serviceInfo.text)
    end
end

function ServiceVehicle.getIsVehicleValid(vehicle, vehicleToCheck)
    if vehicleToCheck ~= nil and vehicleToCheck ~= vehicle then
        return vehicleToCheck.configFileName ~= nil and vehicleToCheck.getActiveFarm ~= nil
    end

    return false
end


ServiceVehicleActivatable = {}
local ServiceVehicleActivatable_mt = Class(ServiceVehicleActivatable)

function ServiceVehicleActivatable:new(vehicle, triggers)
    if vehicle ~= nil and triggers ~= nil then
        local self = {}
        setmetatable(self, ServiceVehicleActivatable_mt)

        self.vehicle = vehicle

        self.playerTrigger = triggers.playerTrigger
        self.sellTriggerNode = triggers.sellTriggerNode
        addTrigger(self.playerTrigger, "triggerCallback", self)
        addTrigger(self.sellTriggerNode, "sellAreaTriggerCallback", self)

        self.playerInTrigger = nil
        self.objectActivated = false

        -- WorkShopTabber Fix
        if VehicleSellingPoint.showNextVehicle ~= nil then
            self.showNextVehicle = VehicleSellingPoint.showNextVehicle
        end

        self.originalOnActivateObject = VehicleSellingPoint.onActivateObject
        self.ownWorkshop = true

        self.vehicleInRange = {}
        self.currentVehicle = nil
        self.currentVehicleId = 0

        self.activateText = "Interact"

        self.hasExtraInputs = false
        self.inputsActive = false
        self.hasMarkers = false

        return self
    else
        return nil
    end
end

function ServiceVehicleActivatable:delete()
    if self.playerTrigger ~= nil then
        removeTrigger(self.playerTrigger)
        self.playerTrigger = nil
    end

    if self.sellTriggerNode ~= nil then
        removeTrigger(self.sellTriggerNode)
        self.sellTriggerNode = nil
    end

    if self.objectActivated then
        g_currentMission:removeActivatableObject(self)
        self.objectActivated = false
        self.playerInTrigger = nil
    end

    self:addRemoveActionEvent()
end

function ServiceVehicleActivatable:getIsActivatable()
    if g_currentMission.controlPlayer and g_currentMission.accessHandler:canFarmAccess(g_currentMission:getFarmId(), self.vehicle) then
        self:setActivatableText()

        return true
    end

    if self.objectActivated then
        g_currentMission:removeActivatableObject(self)
        self.objectActivated = false
        self.playerInTrigger = nil

        if self.hasExtraInputs then
            self:addRemoveActionEvent()
        end
    end

    return false
end

function ServiceVehicleActivatable:onActivateObject()
    if self.vehicle ~= nil then
        if self.hasExtraInputs then
            local spec = self.vehicle.spec_serviceVehicle
            self.vehicle:setServiceAnimationState(not spec.serviceAnimation.isActive, false)
        else
            self:originalOnActivateObject()
        end
    end
end

function ServiceVehicleActivatable:shouldRemoveActivatable()
    return false
end

function ServiceVehicleActivatable:setActivatableText()
    local spec = self.vehicle.spec_serviceVehicle

    if self.hasExtraInputs then
        self:addRemoveActionEvent()

        if spec.serviceAnimation.isActive then
            self.activateText = spec.texts.deactivateService
        else
            self.activateText = spec.texts.activateService
        end
    else
        self.activateText = spec.texts.activateServiceMenu
    end
end

function ServiceVehicleActivatable:addRemoveActionEvent()
    if self.objectActivated then
        if not self.inputsActive then
            g_inputBinding:beginActionEventsModification(Player.INPUT_CONTEXT_NAME)

            if self.eventIdActivateService == nil then
                local _, eventId = g_inputBinding:registerActionEvent(InputAction.SV_ACTIVATE_SERVICE, self, self.onInputActivateService, false, true, false, false)
                self.eventIdActivateService = eventId
            end

            if self.eventIdActivateMarkers == nil and self.hasMarkers then
                local _, eventId = g_inputBinding:registerActionEvent(InputAction.SV_SET_MARKERS, self, self.onInputSetMarkers, false, true, false, false)
                self.eventIdActivateMarkers = eventId
            end

            g_inputBinding:endActionEventsModification()
            self.inputsActive = true
        end
    else
        if self.inputsActive then
            g_inputBinding:beginActionEventsModification(Player.INPUT_CONTEXT_NAME)

            if self.eventIdActivateService ~= nil then
                g_inputBinding:removeActionEvent(self.eventIdActivateService)
                self.eventIdActivateService = nil
            end

            if self.eventIdActivateMarkers ~= nil then
                g_inputBinding:removeActionEvent(self.eventIdActivateMarkers)
                self.eventIdActivateMarkers = nil
            end

            g_inputBinding:endActionEventsModification()
            self.inputsActive = false
        end
    end
end

function ServiceVehicleActivatable:drawActivate()
    if self.hasExtraInputs and self.vehicle ~= nil then
        local spec = self.vehicle.spec_serviceVehicle

        if spec.serviceAnimation ~= nil then
            local active = spec.serviceAnimation.finished

            if self.eventIdActivateService ~= nil then
                g_inputBinding:setActionEventText(self.eventIdActivateService, spec.texts.activateServiceMenu)

                g_inputBinding:setActionEventTextVisibility(self.eventIdActivateService, active)
                g_inputBinding:setActionEventActive(self.eventIdActivateService, active)
            end

            if self.eventIdActivateMarkers ~= nil and self.hasMarkers then
                if spec.serviceAnimation.markers.showMarkers then
                    g_inputBinding:setActionEventText(self.eventIdActivateMarkers, string.format(g_i18n:getText("action_turnOffOBJECT"), spec.texts.triggerMarkers))
                else
                    g_inputBinding:setActionEventText(self.eventIdActivateMarkers, string.format(g_i18n:getText("action_turnOnOBJECT"), spec.texts.triggerMarkers))
                end

                g_inputBinding:setActionEventTextVisibility(self.eventIdActivateMarkers, active)
                g_inputBinding:setActionEventActive(self.eventIdActivateMarkers, active)
            end
        end
    end
end

function ServiceVehicleActivatable.onInputActivateService(self, actionName, inputValue, callbackState, isAnalog)
    if self.objectActivated then
        self:originalOnActivateObject()
    end
end

function ServiceVehicleActivatable.onInputSetMarkers(self, actionName, inputValue, callbackState, isAnalog)
    if self.objectActivated and self.vehicle ~= nil then
        local spec = self.vehicle.spec_serviceVehicle

        if spec.serviceAnimation ~= nil and spec.serviceAnimation.markers ~= nil then
            self.vehicle:setServiceMarkerState(not spec.serviceAnimation.markers.showMarkers, true, false)
        end
    end
end

function ServiceVehicleActivatable:triggerCallback(triggerId, otherId, onEnter, onLeave, onStay)
    if onEnter or onLeave then
        if g_currentMission.player ~= nil and otherId == g_currentMission.player.rootNode then
            if onEnter then
                if not self.objectActivated then
                    g_currentMission:addActivatableObject(self)
                    self.objectActivated = true
                    self.playerInTrigger = g_currentMission.player
                end
            else
                if self.objectActivated then
                    g_currentMission:removeActivatableObject(self)
                    self.objectActivated = false
                    self.playerInTrigger = nil

                    if self.hasExtraInputs then
                        self:addRemoveActionEvent()
                    end
                end
            end
        end
    end
end

function ServiceVehicleActivatable:sellAreaTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
    if otherShapeId ~= nil and (onEnter or onLeave) then
        if onEnter then
            self.vehicleInRange[otherShapeId] = true
        elseif onLeave then
            self.vehicleInRange[otherShapeId] = nil
        end

        self:determineCurrentVehicle()
    end
end

function ServiceVehicleActivatable:determineCurrentVehicle()
    self.currentVehicle = nil

    for vehicleId, inRange in pairs(self.vehicleInRange) do
        if inRange ~= nil then
            local currentVehicle = g_currentMission.nodeToObject[vehicleId]

            if ServiceVehicle.getIsVehicleValid(self.vehicle, currentVehicle) then
                self.currentVehicle = currentVehicle

                if self.currentVehicle:getActiveFarm() ~= self.vehicle:getActiveFarm() or not SpecializationUtil.hasSpecialization(Rideable, self.currentVehicle.specializations) then
                    break
                end

                -- Only if 3D text is used.
                if self.vehicle.spec_serviceVehicle.serviceInformation ~= nil then
                    self.vehicle:raiseActive()
                end
            end
        end

        self.vehicleInRange[vehicleId] = nil
    end
end


ServiceVehicleFuelActivatable = {}
local ServiceVehicleFuelActivatable_mt = Class(ServiceVehicleFuelActivatable)

function ServiceVehicleFuelActivatable:new(vehicle, playerTrigger)
    if vehicle ~= nil and playerTrigger ~= nil then
        local self = {}
        setmetatable(self, ServiceVehicleFuelActivatable_mt)

        self.vehicle = vehicle

        self.playerTrigger = playerTrigger
        addTrigger(self.playerTrigger, "triggerCallback", self)

        self.objectActivated = false
        self.activateText = "Interact"

        self.hasMarkers = false
        self.inputsActive = false

        return self
    end

    return nil
end

function ServiceVehicleFuelActivatable:delete()
    if self.playerTrigger ~= nil then
        removeTrigger(self.playerTrigger)
        self.playerTrigger = nil
    end

    if self.objectActivated then
        g_currentMission:removeActivatableObject(self)
        self.objectActivated = false
    end

    self:addRemoveActionEvent()
end

function ServiceVehicleFuelActivatable:triggerCallback(triggerId, otherId, onEnter, onLeave, onStay)
    if onEnter or onLeave then
        if g_currentMission.player ~= nil and otherId == g_currentMission.player.rootNode then
            if onEnter then
                if not self.objectActivated then
                    g_currentMission:addActivatableObject(self)
                    self.objectActivated = true
                end
            else
                if self.objectActivated then
                    g_currentMission:removeActivatableObject(self)
                    self.objectActivated = false

                    if self.hasMarkers then
                        self:addRemoveActionEvent()
                    end
                end
            end
        end
    end
end

function ServiceVehicleFuelActivatable:getIsActivatable()
    if g_currentMission.controlPlayer and g_currentMission.accessHandler:canFarmAccess(g_currentMission:getFarmId(), self.vehicle) then
        self:setActivatableText()

        return true
    end

    if self.objectActivated then
        g_currentMission:removeActivatableObject(self)
        self.objectActivated = false

        if self.hasMarkers then
            self:addRemoveActionEvent()
        end
    end

    return false
end

function ServiceVehicleFuelActivatable:onActivateObject()
    if self.vehicle ~= nil then
        local spec = self.vehicle.spec_serviceVehicle
        local fillUnit = self.vehicle:getFillUnitByIndex(spec.fuel.fillUnitIndex)

        if fillUnit ~= nil then
            if fillUnit.fillLevel > 0 and fillUnit.capacity ~= 0 then
                self.vehicle:setServiceFuelState(not spec.fuel.isActive, false)
            else
                g_currentMission:showBlinkingWarning(spec.texts.fuelTankEmpty)
            end
        end
    end
end

function ServiceVehicleFuelActivatable:drawActivate()
    if self.vehicle ~= nil and self.hasMarkers and self.eventIdActivateMarkers ~= nil then
        local spec = self.vehicle.spec_serviceVehicle

        if spec.fuel.markers.showMarkers then
            g_inputBinding:setActionEventText(self.eventIdActivateMarkers, string.format(g_i18n:getText("action_turnOffOBJECT"), spec.texts.triggerMarkers))
        else
            g_inputBinding:setActionEventText(self.eventIdActivateMarkers, string.format(g_i18n:getText("action_turnOnOBJECT"), spec.texts.triggerMarkers))
        end

        g_inputBinding:setActionEventTextVisibility(self.eventIdActivateMarkers, spec.fuel.isActive)
        g_inputBinding:setActionEventActive(self.eventIdActivateMarkers, spec.fuel.isActive)
    end
end

function ServiceVehicleFuelActivatable:shouldRemoveActivatable()
    return false
end

function ServiceVehicleFuelActivatable:setActivatableText()
    local spec = self.vehicle.spec_serviceVehicle
    local percent = 100 * (self.vehicle:getFillUnitFillLevelPercentage(spec.fuel.fillUnitIndex) or 0)

    if spec.fuel.isActive then
        self.activateText = string.format("%s ( %s - %d %% )", spec.texts.stopFuelPump, spec.texts.diesel, percent)
    else
        self.activateText = string.format("%s ( %s - %d %% )", spec.texts.startFuelPump, spec.texts.diesel, percent)
    end

    self:addRemoveActionEvent()
end

function ServiceVehicleFuelActivatable:addRemoveActionEvent()
    if self.objectActivated then
        if not self.inputsActive then
            if self.eventIdActivateMarkers == nil and self.hasMarkers then
                g_inputBinding:beginActionEventsModification(Player.INPUT_CONTEXT_NAME)

                local _, eventId = g_inputBinding:registerActionEvent(InputAction.SV_SET_MARKERS, self, self.onInputSetMarkers, false, true, false, false)
                self.eventIdActivateMarkers = eventId

                g_inputBinding:endActionEventsModification()
            end

            self.inputsActive = true
        end
    else
        if self.eventIdActivateMarkers ~= nil then
            g_inputBinding:beginActionEventsModification(Player.INPUT_CONTEXT_NAME)

            g_inputBinding:removeActionEvent(self.eventIdActivateMarkers)
            self.eventIdActivateMarkers = nil

            g_inputBinding:endActionEventsModification()
        end

        self.inputsActive = false
    end
end

function ServiceVehicleFuelActivatable.onInputSetMarkers(self, actionName, inputValue, callbackState, isAnalog)
    if self.objectActivated and self.vehicle ~= nil then
        local fuelMarkers = self.vehicle.spec_serviceVehicle.fuel.markers
        self.vehicle:setServiceMarkerState(not fuelMarkers.showMarkers, false, false)
    end
end


ServiceVehicleChargingActivatable = {}
local ServiceVehicleChargingActivatable_mt = Class(ServiceVehicleChargingActivatable)

function ServiceVehicleChargingActivatable:new(vehicle)
    local spec = vehicle.spec_serviceVehicle

    if spec ~= nil and spec.fuel ~= nil then
        local self = {}
        setmetatable(self, ServiceVehicleChargingActivatable_mt)

        self.vehicle = vehicle

        self.stopChargingText = ServiceVehicle.returnText("action_stopCharging", "Stop Charging", "pdlc_alpineFarmingPack")
        self.startChargingText = ServiceVehicle.returnText("action_startCharging", "Start Charging", "pdlc_alpineFarmingPack")

        self.activateText = self.startChargingText

        return self
    end

    return nil
end

function ServiceVehicleChargingActivatable:delete()
    if self.vehicle ~= nil then
        self.vehicle.spec_serviceVehicle.electricVehicleInTrigger = false
    end

    g_currentMission:removeActivatableObject(self)
end

function ServiceVehicleChargingActivatable:getTargetVehicle()
    if g_currentMission.controlledVehicle ~= nil then
        local spec = self.vehicle.spec_serviceVehicle
        local isChargingVehicle = spec.isChargingVehicle and spec.currentElectricVehicle ~= nil

        for targetVehicle, fillUnitIndex in pairs (spec.electricVehiclesInRange) do
            if targetVehicle:getRootVehicle() == g_currentMission.controlledVehicle then
                if isChargingVehicle then
                    if spec.currentElectricVehicle == targetVehicle then
                        return targetVehicle, fillUnitIndex, true, true
                    end
                else
                    local fillUnit = targetVehicle:getFillUnitByIndex(fillUnitIndex)
                    return targetVehicle, fillUnitIndex, false, fillUnit ~= nil and (fillUnit.fillLevel < fillUnit.capacity - 1)
                end
            end
        end
    end

    return nil
end

function ServiceVehicleChargingActivatable:getIsActivatable()
    if self.vehicle ~= nil then
        local targetVehicle, _, isCharging, hasSpace = self:getTargetVehicle()

        if targetVehicle ~= nil and hasSpace then
            if isCharging then
                self.activateText = self.stopChargingText
            else
                self.activateText = self.startChargingText
            end

            return true
        end
    else
        self:delete()
    end

    return false
end

function ServiceVehicleChargingActivatable:onActivateObject()
    if self.vehicle ~= nil then
        if self.vehicle:getIsChargingVehicle() then
            self.vehicle:setIsCharging(false)
        else
            local targetVehicle, fillUnitIndex, _, hasSpace = self:getTargetVehicle()

            if targetVehicle ~= nil and hasSpace then
                self.vehicle:setIsCharging(fillUnitIndex ~= nil, targetVehicle, fillUnitIndex)
            end
        end

        g_currentMission:addActivatableObject(self)
    end
end

function ServiceVehicleChargingActivatable:drawActivate()
    return
end


ServiceVehicleSetIsChargingEvent = {}
local ServiceVehicleSetIsChargingEvent_mt = Class(ServiceVehicleSetIsChargingEvent, Event)
InitEventClass(ServiceVehicleSetIsChargingEvent, "ServiceVehicleSetIsChargingEvent")

function ServiceVehicleSetIsChargingEvent:emptyNew()
    local self = Event:new(ServiceVehicleSetIsChargingEvent_mt)

    return self
end

function ServiceVehicleSetIsChargingEvent:new(object, isCharging, targetVehicle, fillUnitIndex)
    local self = ServiceVehicleSetIsChargingEvent:emptyNew()

    self.object = object
    self.isCharging = isCharging
    self.targetVehicle = targetVehicle
    self.fillUnitIndex = fillUnitIndex

    return self
end

function ServiceVehicleSetIsChargingEvent:readStream(streamId, connection)
    self.object = NetworkUtil.readNodeObject(streamId)
    self.isCharging = streamReadBool(streamId)

    if self.isCharging then
        self.targetVehicle = NetworkUtil.readNodeObject(streamId)
        self.fillUnitIndex = streamReadUInt8(streamId)
    end

    self:run(connection)
end

function ServiceVehicleSetIsChargingEvent:writeStream(streamId, connection)
    NetworkUtil.writeNodeObject(streamId, self.object)

    if streamWriteBool(streamId, self.isCharging) then
        NetworkUtil.writeNodeObject(streamId, self.targetVehicle)
        streamWriteUInt8(streamId, self.fillUnitIndex)
    end
end

function ServiceVehicleSetIsChargingEvent:run(connection)
    self.object:setIsCharging(self.isCharging, self.targetVehicle, self.fillUnitIndex, true)

    if not connection:getIsServer() then
        g_server:broadcastEvent(ServiceVehicleSetIsChargingEvent:new(self.object, self.isCharging, self.targetVehicle, self.fillUnitIndex), nil, connection, self.object)
    end
end

function ServiceVehicleSetIsChargingEvent.sendEvent(object, isCharging, targetVehicle, fillUnitIndex, noEventSend)
    if isCharging ~= object.isCharging and (noEventSend == nil or noEventSend == false) then
        if g_server ~= nil then
            g_server:broadcastEvent(ServiceVehicleSetIsChargingEvent:new(object, isCharging, targetVehicle, fillUnitIndex), nil, nil, object)
        else
            g_client:getServerConnection():sendEvent(ServiceVehicleSetIsChargingEvent:new(object, isCharging, targetVehicle, fillUnitIndex))
        end
    end
end


ServiceVehicleAnimationEvent = {}
local ServiceVehicleAnimationEvent_mt = Class(ServiceVehicleAnimationEvent, Event)
InitEventClass(ServiceVehicleAnimationEvent, "ServiceVehicleAnimationEvent")

function ServiceVehicleAnimationEvent:emptyNew()
    local self = Event:new(ServiceVehicleAnimationEvent_mt)
    return self
end

function ServiceVehicleAnimationEvent:new(object, isActive)
    local self = ServiceVehicleAnimationEvent:emptyNew()
    self.object = object
    self.isActive = isActive

    return self
end

function ServiceVehicleAnimationEvent:readStream(streamId, connection)
    self.object = NetworkUtil.readNodeObject(streamId)
    self.isActive = streamReadBool(streamId)

    self:run(connection)
end

function ServiceVehicleAnimationEvent:writeStream(streamId, connection)
    NetworkUtil.writeNodeObject(streamId, self.object)
    streamWriteBool(streamId, self.isActive)
end

function ServiceVehicleAnimationEvent:run(connection)
    self.object:setServiceAnimationState(self.isActive, true)

    if not connection:getIsServer() then
        g_server:broadcastEvent(ServiceVehicleAnimationEvent:new(self.object, self.isActive), nil, connection, self.object)
    end
end

function ServiceVehicleAnimationEvent.sendEvent(vehicle, isActive, noEventSend)
    if vehicle.spec_serviceVehicle ~= nil and vehicle.spec_serviceVehicle.serviceAnimation ~= nil and
        vehicle.spec_serviceVehicle.serviceAnimation.isActive ~= isActive then
        if noEventSend == nil or noEventSend == false then
            if g_server ~= nil then
                g_server:broadcastEvent(ServiceVehicleAnimationEvent:new(vehicle, isActive), nil, nil, vehicle)
            else
                g_client:getServerConnection():sendEvent(ServiceVehicleAnimationEvent:new(vehicle, isActive))
            end
        end
    end
end


ServiceVehicleFuelEvent = {}
local ServiceVehicleFuelEvent_mt = Class(ServiceVehicleFuelEvent, Event)
InitEventClass(ServiceVehicleFuelEvent, "ServiceVehicleFuelEvent")

function ServiceVehicleFuelEvent:emptyNew()
    local self = Event:new(ServiceVehicleFuelEvent_mt)
    return self
end

function ServiceVehicleFuelEvent:new(object, isActive)
    local self = ServiceVehicleFuelEvent:emptyNew()
    self.object = object
    self.isActive = isActive

    return self
end

function ServiceVehicleFuelEvent:readStream(streamId, connection)
    self.object = NetworkUtil.readNodeObject(streamId)
    self.isActive = streamReadBool(streamId)

    self:run(connection)
end

function ServiceVehicleFuelEvent:writeStream(streamId, connection)
    NetworkUtil.writeNodeObject(streamId, self.object)
    streamWriteBool(streamId, self.isActive)
end

function ServiceVehicleFuelEvent:run(connection)
    self.object:setServiceFuelState(self.isActive, true)

    if not connection:getIsServer() then
        g_server:broadcastEvent(ServiceVehicleFuelEvent:new(self.object, self.isActive), nil, connection, self.object)
    end
end

function ServiceVehicleFuelEvent.sendEvent(vehicle, isActive, noEventSend)
    if vehicle.spec_serviceVehicle ~= nil and vehicle.spec_serviceVehicle.fuel ~= nil and
        vehicle.spec_serviceVehicle.fuel.isActive ~= isActive then
        if noEventSend == nil or noEventSend == false then
            if g_server ~= nil then
                g_server:broadcastEvent(ServiceVehicleFuelEvent:new(vehicle, isActive), nil, nil, vehicle)
            else
                g_client:getServerConnection():sendEvent(ServiceVehicleFuelEvent:new(vehicle, isActive))
            end
        end
    end
end


ServiceVehicleMarkerEvent = {}
local ServiceVehicleMarkerEvent_mt = Class(ServiceVehicleMarkerEvent, Event)
InitEventClass(ServiceVehicleMarkerEvent, "ServiceVehicleMarkerEvent")

function ServiceVehicleMarkerEvent:emptyNew()
    local self = Event:new(ServiceVehicleMarkerEvent_mt)
    return self
end

function ServiceVehicleMarkerEvent:new(object, showMarkers, isService)
    local self = ServiceVehicleMarkerEvent:emptyNew()
    self.object = object
    self.showMarkers = showMarkers
    self.isService = isService

    return self
end

function ServiceVehicleMarkerEvent:readStream(streamId, connection)
    self.object = NetworkUtil.readNodeObject(streamId)
    self.showMarkers = streamReadBool(streamId)
    self.isService = streamReadBool(streamId)

    self:run(connection)
end

function ServiceVehicleMarkerEvent:writeStream(streamId, connection)
    NetworkUtil.writeNodeObject(streamId, self.object)
    streamWriteBool(streamId, self.showMarkers)
    streamWriteBool(streamId, self.isService)
end

function ServiceVehicleMarkerEvent:run(connection)
    self.object:setServiceMarkerState(self.showMarkers, self.isService, true)

    if not connection:getIsServer() then
        g_server:broadcastEvent(ServiceVehicleMarkerEvent:new(self.object, self.showMarkers, self.isService), nil, connection, self.object)
    end
end

function ServiceVehicleMarkerEvent.sendEvent(vehicle, showMarkers, isService, noEventSend)
    if noEventSend == nil or noEventSend == false then
        if g_server ~= nil then
            g_server:broadcastEvent(ServiceVehicleMarkerEvent:new(vehicle, showMarkers, isService), nil, nil, vehicle)
        else
            g_client:getServerConnection():sendEvent(ServiceVehicleMarkerEvent:new(vehicle, showMarkers, isService))
        end
    end
end
